"""Exceptions for ilamb."""


class ILAMBException(Exception):
    """Exceptions from the intake-esgf package."""


class MissingVariable(ILAMBException):
    """A required variable is missing."""


class VarNotInModel(ILAMBException):
    """This variable is not in the model at all."""


class TemporalOverlapIssue(ILAMBException):
    """This variable does not have the correct temporal overal to be comparable."""


class MissingRegion(ILAMBException):
    """You are trying to use a region that is not registered."""


class NoDatabaseEntry(ILAMBException):
    """The quantile database does not contain that which you are searching."""


class NoSiteDimension(ILAMBException):
    """The dataset/dataarray does not contain a clear site dimension."""


class NoUncertainty(ILAMBException):
    """The dataset/dataarray does not contain uncertainty."""


class AnalysisNotAppropriate(ILAMBException):
    """The dataset/dataarray is not appropriate for an analysis"""


class AnalysisFailure(ILAMBException):
    """
    The analysis function you were running threw an exception.

    Parameters
    ----------
    analysis : str
        The name of the analysis which has failed.
    model : str
        The name of the model whose analysis failed.
    """

    def __init__(self, analysis: str, model: str):  # numpydoc ignore=GL08
        self.analysis = analysis
        self.model = model

    def __str__(self):  # numpydoc ignore=GL08
        return f"The '{self.analysis}' analysis failed for '{self.model}'"



from ilamb3.transform.amoc import msftmz_to_rapid
from ilamb3.transform.gradient import depth_gradient
from ilamb3.transform.ohc import ocean_heat_content
from ilamb3.transform.select import select_depth
from ilamb3.transform.soilmoisture import soil_moisture_to_vol_fraction
from ilamb3.transform.stratification_index import stratification_index

ALL_TRANSFORMS = {
    "soil_moisture_to_vol_fraction": soil_moisture_to_vol_fraction,
    "msftmz_to_rapid": msftmz_to_rapid,
    "stratification_index": stratification_index,
    "ocean_heat_content": ocean_heat_content,
    "select_depth": select_depth,
    "depth_gradient": depth_gradient,
}
__all__ = list(ALL_TRANSFORMS.keys())



import matplotlib.pyplot as plt
import numpy as np

from ilamb3.analysis import hydro_analysis
from ilamb3.tests.test_run import generate_test_dset


def test_hydro():
    ref = generate_test_dset("tas", "degC", nyear=2, nlat=2, nlon=3, seed=1)
    com = generate_test_dset("tas", "degC", nyear=2, nlat=2, nlon=3, seed=2)
    analysis = hydro_analysis("tas")
    assert set(["tas"]) == set(analysis.required_variables())
    ds_com = {}
    df, ds_ref, ds_com["Comparison"] = analysis(ref, com)
    # check scalars
    assert len(df) == 35
    q = df[(df["analysis"] == "Fall (SON)") & (df["type"] == "score")]
    assert len(q) == 1
    assert np.allclose(q.iloc[0].value, 0.6474532320617079)
    plt.rcParams.update({"figure.max_open_warning": 0})
    dfp = analysis.plots(df, ds_ref, ds_com)
    assert len(dfp) == 35
    plt.close()



"""
An abstract class for implementing analysis functions used in ILAMB.
"""

from abc import ABC, abstractmethod
from typing import Any

import pandas as pd
import xarray as xr

import ilamb3.dataset as dset
import ilamb3.regions as ilr


class ILAMBAnalysis(ABC):
    """
    The ILAMB analysis base class.

    An abstract base class (ABC) in python is a way to define the structure of
    an object that can be used in other parts of the system. In our case this
    means that in order for your analysis to be compatible in the ILAMB system,
    you need to write a class that has *at minimum* the following member
    functions with their arguments and return values.
    """

    @abstractmethod
    def __init__(self, **kwargs: Any):
        """
        Initialize the analysis.

        This is run when you initialize the analysis. Any options that your
        analysis method will require should be input and stored here as
        keywords.
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def required_variables(self) -> list[str] | dict[str, list[str]]:
        """
        Return the variables used in this analysis.

        Your analysis must be able to provide a list of variables that are to be
        used. This is so we can query models with your function and learn which
        variables are to be used. The assumption is that the variables you
        require are part of a historical-like experiment. If your analysis
        requires variables from different experiments, this function can also
        return a dictionary of lists where they keys are the experiment names
        from which the lists of variables are required.
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def __call__(
        self, ref: xr.Dataset, com: xr.Dataset
    ) -> tuple[pd.DataFrame, xr.Dataset, xr.Dataset]:
        """
        The function which performs the analysis.

        The work of your analysis must accept a reference and comparison
        dataset. The user will need to package everything that each needs for
        the analysis upon submission. Inside this function you can use whatever
        you want to generate scalars and maps/curves for plotting. Once
        complete, ILAMB would like results in a pandas dataframe (for scalar
        information) as well as a dataset for the reference and comparison.

        Parameters
        ----------
        ref : xr.Dataset
            The reference dataset.
        com : xr.Dataset
            The comparison dataset.

        Returns
        -------
        pd.DataFrame
            The scalars and scores to be returned.
        xr.Dataset
            The reference intermediate maps and curves to plot
        xr.Dataset
            The comparison intermediate maps and curves to plot
        """
        raise NotImplementedError()  # pragma: no cover

    @abstractmethod
    def plots(
        self, df: pd.DataFrame, ref: xr.Dataset, com: dict[str, xr.Dataset]
    ) -> pd.DataFrame:
        """
        The function which creates plots from analysis intermediate results.

        ILAMB will run your __call__ method over all comparisons and build up a
        dataframe of scalars and a dictionary of intermediate result datasets.
        We will pass these into this function where you can generate plots.
        Instead of saving them directly here, place the into a pandas dataframe
        which is returned to the user.

        Parameters
        ----------
        df : pd.DataFrame
            The scalars and scores from all comparisons.
        ref : xr.Dataset
            The reference intermediate maps and curves to plot
        com : dict[str,xr.Dataset]
            A dictionary of comparison intermediate maps and curves to plot,
            whose keys are the model names.

        Returns
        -------
        pd.DataFrame
            The dataframe of plots.
        """
        raise NotImplementedError()  # pragma: no cover


def add_overall_score(df: pd.DataFrame) -> pd.DataFrame:
    """
    Synthesize an average 'Overall' score from all scores in the dataframe.
    """
    add = (
        df[df["type"] == "score"]
        .groupby(["source", "region"])
        .mean(numeric_only=True)
        .reset_index()
    )
    add["name"] = "Overall Score [1]"
    add["type"] = "score"
    add["units"] = "1"
    df = pd.concat([df, add]).reset_index(drop=True)
    return df


def integrate_or_mean(
    var: xr.DataArray | xr.Dataset, varname: str, region: str | None, mean: bool
) -> xr.DataArray:
    """
    Integration/average the input dataarray/dataset to reduce in space/site.
    """
    da = var
    if isinstance(var, xr.Dataset):
        da = var[varname]
    if dset.is_spatial(da):
        da = dset.integrate_space(
            da,
            varname,
            region=region,
            mean=mean,
        )
    elif dset.is_site(da):
        da = ilr.Regions().restrict_to_region(da, region)
        da = da.mean(dim=dset.get_dim_name(da, "site"))
    else:
        raise ValueError(f"Input is neither spatial nor site: {da}")
    return da


def scalarify(
    var: xr.DataArray | xr.Dataset, varname: str, region: str | None, mean: bool
) -> tuple[float, str]:
    """
    Integration/average the input dataarray/dataset to generate a scalar.
    """
    da = integrate_or_mean(var, varname, region, mean)
    da = da.pint.quantify()
    return float(da.pint.dequantify()), f"{da.pint.units:~cf}"



"""
The ILAMB RMSE methodology.

See Also
--------
ILAMBAnalysis : The abstract base class from which this derives.
"""

from typing import Any, Literal

import numpy as np
import pandas as pd
import xarray as xr

import ilamb3.plot as plt
from ilamb3 import compare as cmp
from ilamb3 import dataset as dset
from ilamb3.analysis.base import ILAMBAnalysis, integrate_or_mean, scalarify
from ilamb3.exceptions import AnalysisNotAppropriate, NoUncertainty


class rmse_analysis(ILAMBAnalysis):
    """
    The ILAMB RMSE methodology.

    Parameters
    ----------
    required_variable : str
        The name of the variable to be used in this analysis.
    method : str
        The name of the scoring methodology to use, either `Collier2018` or
        `RegionalQuantiles`.
    regions : list
        A list of region labels over which to apply the analysis.
    use_uncertainty : bool
        Enable to utilize uncertainty information from the reference product if
        present.
    quantile_dbase : pd.DataFrame
        If using `method='RegionalQuantiles'`, the dataframe containing the
        regional quantiles to be used to score the datasets.
    quantile_threshold : int
        If using `method='RegionalQuantiles'`, the threshold values to use from
        the `quantile_dbase`.

    Methods
    -------
    required_variables
        What variables are required.
    __call__
        The method
    """

    def __init__(
        self,
        required_variable: str,
        score_basis: Literal["series", "cycle"] = "series",
        regions: list[str | None] = [None],
        use_uncertainty: bool = True,
        **kwargs: Any,  # this is so we can pass extra arguments without failure
    ):
        self.req_variable = required_variable
        self.score_basis = score_basis
        self.regions = regions
        self.use_uncertainty = use_uncertainty
        self.kwargs = kwargs

    def required_variables(self) -> list[str]:
        """
        Return the list of variables required for this analysis.

        Returns
        -------
        list
            The variable names used in this analysis.
        """
        return [self.req_variable]

    def __call__(
        self,
        ref: xr.Dataset,
        com: xr.Dataset,
    ) -> tuple[pd.DataFrame, xr.Dataset, xr.Dataset]:
        """
        Apply the ILAMB RMSE methodology on the given datasets.

        Parameters
        ----------
        ref : xr.Dataset
            The reference dataset.
        com : xr.Dataset
            The comparison dataset.

        Returns
        -------
        pd.DataFrame
            A dataframe with scalar and score information from the comparison.
        xr.Dataset
            A dataset containing reference grided information from the comparison.
        xr.Dataset
            A dataset containing comparison grided information from the comparison.
        """
        # Initialize
        ANALYSIS_NAME = "RMSE"
        varname = self.req_variable

        # Make the variables comparable and force loading into memory
        ref, com = cmp.make_comparable(ref, com, varname)

        # Is the time series long enough for this to be meaningful?
        if len(ref[dset.get_dim_name(ref, "time")]) < 24:
            raise AnalysisNotAppropriate()
        if len(com[dset.get_dim_name(com, "time")]) < 24:
            raise AnalysisNotAppropriate()

        # Before operating on these, compute spatial means
        ds_ref = {}
        ds_com = {}
        for region in self.regions:
            ds_ref[f"trace_{region}"] = integrate_or_mean(
                ref, varname, region, mean=True
            )
            ds_com[f"trace_{region}"] = integrate_or_mean(
                com, varname, region, mean=True
            )

        # Move calendars
        ref = cmp.convert_calendar_monthly_noleap(ref)
        com = cmp.convert_calendar_monthly_noleap(com)

        # Get the reference data uncertainty, only use if present and desired
        args = [ref, com]
        if self.use_uncertainty:
            try:
                uncert = dset.get_scalar_uncertainty(ref, varname)
                args.append(uncert)
            except NoUncertainty:
                self.use_uncertainty = False
                uncert = None

        # Conversions
        if self.use_uncertainty:
            ref, com, uncert = cmp.rename_dims(
                cmp.nest_spatial_grids(ref[varname], com[varname], uncert)
            )
        else:
            ref, com = cmp.rename_dims(
                *cmp.nest_spatial_grids(ref[varname], com[varname])
            )

        # Compute the RMSE and score
        rmse = np.sqrt(dset.integrate_time((com - ref) ** 2, varname, mean=True))
        ref_mean = dset.integrate_time(ref, varname, mean=True)
        com_mean = dset.integrate_time(com, varname, mean=True)
        crmse = np.sqrt(
            dset.integrate_time(
                (
                    np.abs((com - com_mean) - (ref - ref_mean))
                    - (uncert if self.use_uncertainty else 0.0)
                ).clip(min=0)
                ** 2,
                varname,
                mean=True,
            )
        )
        crms = np.sqrt(dset.integrate_time((ref - ref_mean) ** 2, varname, mean=True))
        score = np.exp(-crmse / crms)
        score.attrs["units"] = "1"

        # Load outputs and scalars
        ds_com["rmse"] = rmse
        ds_com["rmsescore"] = score
        df = []
        for region in self.regions:
            val, unit = scalarify(rmse, varname, region, mean=True)
            df += [
                {
                    "source": "Comparison",
                    "region": str(region),
                    "analysis": ANALYSIS_NAME,
                    "name": "RMSE",
                    "type": "scalar",
                    "units": unit,
                    "value": val,
                },
            ]
            val, _ = scalarify(score, varname, region, mean=True)
            df += [
                {
                    "source": "Comparison",
                    "region": str(region),
                    "analysis": ANALYSIS_NAME,
                    "name": "RMSE Score",
                    "type": "score",
                    "units": "1",
                    "value": val,
                },
            ]

        df = pd.DataFrame(df)
        ds_ref = xr.merge([ds_ref])
        ds_com = xr.merge([ds_com])
        return df, ds_ref, ds_com

    def plots(
        self,
        df: pd.DataFrame,
        ref: xr.Dataset,
        com: dict[str, xr.Dataset],
    ) -> pd.DataFrame:
        # This analysis was not run and we should skip plotting entirely
        if "RMSE" not in df["analysis"].unique():
            return pd.DataFrame()

        # Some initialization
        regions = [None if r == "None" else r for r in df["region"].unique()]
        com["Reference"] = ref

        # Setup plot data
        df = plt.determine_plot_limits(com).set_index("name")
        df.loc["rmse", ["cmap", "title"]] = ["Oranges", "RMSE"]
        df.loc["rmsescore", ["cmap", "title"]] = ["plasma", "RMSE Score"]

        # Build up a dataframe of matplotlib axes
        axs = [
            {
                "name": plot,
                "title": df.loc[plot, "title"],
                "region": region,
                "source": source,
                "axis": (
                    plt.plot_map(
                        ds[plot],
                        region=region,
                        vmin=df.loc[plot, "low"],
                        vmax=df.loc[plot, "high"],
                        cmap=df.loc[plot, "cmap"],
                        title=source + " " + df.loc[plot, "title"],
                    )
                    if plot in ds
                    else pd.NA
                ),
            }
            for plot in ["rmse", "rmsescore"]
            for source, ds in com.items()
            for region in regions
        ]

        axs += [
            {
                "name": "trace",
                "title": "Time Series",
                "region": plot.split("_")[-1],
                "source": source,
                "axis": (
                    plt.plot_curve(
                        {source: ds} | {"Reference": ref},
                        plot,
                        vmin=df.loc[plot, "low"]
                        - 0.05 * (df.loc[plot, "high"] - df.loc[plot, "low"]),
                        vmax=df.loc[plot, "high"]
                        + 0.05 * (df.loc[plot, "high"] - df.loc[plot, "low"]),
                        title=f"{source} Time Series",
                        label="",
                    )
                    if plot in ds
                    else pd.NA
                ),
            }
            for plot in [f"trace_{region}" for region in regions]
            for source, ds in com.items()
            if source != "Reference"
        ]

        axs = pd.DataFrame(axs).dropna(subset=["axis"])
        return axs



import cftime as cf
import numpy as np
import pytest
import xarray as xr

from ilamb3.tests.test_run import generate_test_dset
from ilamb3.transform import ALL_TRANSFORMS

ALL_TRANSFORMS


def gen_msftmz(seed: int = 1):
    rs = np.random.RandomState(seed)
    coords = {}
    # time, basin, lev, lat
    coords["time"] = [
        cf.DatetimeNoLeap(2000 + int(m / 12), (m % 12) + 1, 15) for m in range(1 * 12)
    ]
    coords["basin"] = ["hi", "there"]
    coords["lev"] = 0.5 * (
        np.linspace(0, 100, 10 + 1)[1:] + np.linspace(0, 100, 10 + 1)[:-1]
    )
    coords["lat"] = 0.5 * (
        np.linspace(-90, 90, 10 + 1)[1:] + np.linspace(-90, 90, 10 + 1)[:-1]
    )
    dims = [c for c in coords]
    return xr.Dataset(
        data_vars={
            "msftmz": xr.DataArray(
                rs.rand(*[len(coords[d]) for d in dims]) * 20,
                coords=coords,
                dims=dims,
                name="msftmz",
                attrs={"units": "kg s-1"},
            ),
        }
    )


DATA = {
    "soil_moisture_to_vol_fraction": generate_test_dset(
        "mrsos", "kg m-2", nyear=1, nlat=2, nlon=4
    ),
    "msftmz_to_rapid": gen_msftmz(),
    "ocean_heat_content": xr.merge(
        [
            generate_test_dset(
                "thetao",
                "K",
                nyear=10,
                nlat=2,
                nlon=4,
                ndepth=10,
                scale=20,
                shift=273.0,
            ),
            generate_test_dset("volcello", "m3", nlat=2, nlon=4, ndepth=10, scale=1e10),
        ]
    ),
    "select_depth": generate_test_dset(
        "thetao", "K", nyear=1, nlat=2, nlon=4, ndepth=10
    ),
    "depth_gradient": generate_test_dset(
        "thetao", "K", nyear=1, nlat=2, nlon=4, ndepth=10
    ),
}


@pytest.mark.parametrize(
    "name,kwargs,out,value",
    [
        ("soil_moisture_to_vol_fraction", {}, "mrsos", 0.09702962798695143),
        ("msftmz_to_rapid", {}, "amoc", 1.763701642255525e-08),
        ("ocean_heat_content", {"reference_year": 2000}, "ohc", -0.0046867534613105055),
        ("select_depth", {"value": 0}, "thetao", 9.43861150676275),
        ("select_depth", {"vmin": 1, "vmax": 40}, "thetao", 9.983843647875275),
        ("depth_gradient", {}, "thetao_depth_gradient", -0.005550578833866464),
    ],
)
def test_transform(name, kwargs, out, value):
    transform = ALL_TRANSFORMS[name](**kwargs)
    ds = transform(DATA[name])
    assert np.allclose(value, ds[out].mean().values)



"""
An ILAMB transform for converting integrated soil moisture to a density.
"""

import xarray as xr

import ilamb3.dataset as dset
from ilamb3.transform.base import ILAMBTransform


class soil_moisture_to_vol_fraction(ILAMBTransform):
    """
    Convert depth integrated moisture to a volumetric fraction.
    """

    def __init__(self):
        pass

    def required_variables(self) -> list[str]:
        """
        Return the variables this transform can convert.
        """
        return ["mrsos", "mrsol"]

    def __call__(self, ds: xr.Dataset) -> xr.Dataset:
        """
        Convert depth integrated moisture to a volumnetric fraction.
        """
        for var in self.required_variables():
            if var in ds:
                ds[var] = _to_vol_fraction(ds, var).squeeze()
        return ds


def _to_vol_fraction(ds: xr.Dataset, varname: str) -> xr.DataArray:
    """
    Convert the array from a mass area density to a volume fraction.
    """
    da = ds[varname]
    da = da.pint.quantify()
    ureg = da.pint.registry
    quantity = 1.0 * da.pint.units
    if quantity.check(""):
        return da.pint.dequantify()
    if not quantity.check("[mass] / [length]**2"):
        raise ValueError("Cannot convert array of this type.")
    if dset.is_layered(da):
        depth_name = dset.get_dim_name(da, "depth")
        depth = da[depth_name]
        if "bounds" not in depth.attrs:
            raise ValueError(
                "Cannot convert soil mositure data when depth bounds are not coordinates."
            )
        depth_name = depth.attrs["bounds"]
        if depth_name not in ds:
            raise ValueError(
                "Cannot convert soil mositure data when depth bounds are not coordinates."
            )
        depth = ds[depth_name] * da.pint.registry.meter
        da = da / depth.diff(dim=depth.dims[-1])
    else:
        # If not layered, we are assuming this is mrsos which is the moisture in
        # the top 10cm
        depth = 0.1 * da.pint.registry.meter
        da = da / depth
    water_density = 998.2071 * ureg.kilogram / ureg.meter**3
    da = (da / water_density).pint.to("dimensionless")
    da = da.pint.dequantify()
    return da



import pandas as pd
import pytest

from ilamb3.analysis.quantiles import check_quantile_database, create_quantile_map
from ilamb3.exceptions import MissingRegion, NoDatabaseEntry


def test_check():
    df = None
    with pytest.raises(ValueError):
        check_quantile_database(df)
    df = pd.DataFrame([{"region": "fake_region"}, {"region": "another_fake_region"}])
    with pytest.raises(MissingRegion):
        check_quantile_database(df)


def test_map():
    df = pd.DataFrame(
        [
            {"quantile": 50, "type": "bias", "variable": "var"},
            {"quantile": 50, "type": "rmse", "variable": "var"},
        ]
    )
    with pytest.raises(NoDatabaseEntry):
        create_quantile_map(df, "var", "bias", 70)



"""
An ILAMB transform for estimating the stratification index from thetao and so.
"""

from typing import Any

import numpy as np
import xarray as xr

import ilamb3.dataset as dset
from ilamb3.transform.base import ILAMBTransform


class stratification_index(ILAMBTransform):
    """
    An ILAMB transform for estimating the stratification index over a region.

    Parameters
    ----------
    depth_horizon : float, optional
        The maxmimum depth to consider in the calculation.
    lat_min : float, optional
        The minimum latitude to consider.
    lat_max : float, optional
        The maximum latitude to consider.
    """

    def __init__(
        self,
        depth_horizon: float = 1000,
        lat_min: float = -55,
        lat_max: float = -30,
        **kwargs: Any,
    ):
        self.depth_horizon = depth_horizon
        self.lat_min = lat_min
        self.lat_max = lat_max

    def required_variables(self):
        """
        Return the variables required by this transform.
        """
        return ["thetao", "so"]

    def __call__(self, ds: xr.Dataset) -> xr.DataArray:
        """
        Estimate the stratification index from thetao and so using the `gsw` package.
        """
        if not set(self.required_variables()).issubset(ds):
            return ds

        try:
            import gsw
        except ImportError:
            raise ImportError(
                "The use of the `stratification_index` transform requires additional dependencies. Try `uv sync --group gcb` and then run again."
            )

        # Reads the temperature and salinity and ensure units
        ds["thetao"] = ds["thetao"].pint.quantify().pint.to("degC")
        ds["so"] = ds["so"].pint.quantify().pint.to("psu")
        ds = ds.pint.dequantify()

        # Convenience definitions
        Ts = ds["thetao"]
        Ss = ds["so"]
        lat_name = dset.get_dim_name(ds, "lat")
        lon_name = dset.get_dim_name(ds, "lon")
        depth_name = dset.get_dim_name(ds, "depth")

        # Calculates the sea pressure
        Pp = gsw.p_from_z(
            -xr.ones_like(ds["thetao"]) * ds["depth"],
            xr.ones_like(ds["thetao"]) * ds[lat_name],
        )

        # Calculate Absolute Salinity and Conservative Temperature
        SA = gsw.SA_from_SP(
            Ss,
            Pp,
            xr.ones_like(ds["thetao"]) * ds[lon_name],
            xr.ones_like(ds["thetao"]) * ds[lat_name],
        )

        # Calculate insitu density (surface and 1000m) from Absolute salinity and Conservative temperature
        rho = gsw.rho(SA, gsw.CT_from_pt(SA, Ts), Pp)

        # Calculate the stratification index using density differences
        SO_SI = rho.interp(
            {depth_name: self.depth_horizon}, method="linear"
        ) - rho.isel({depth_name: 0}).sel({lat_name: slice(self.lat_min, self.lat_max)})

        # Area weights
        area_SO_SI = np.cos(np.deg2rad(SO_SI[lat_name]))
        weights_2D = xr.ones_like(SO_SI) * area_SO_SI

        ds["SI"] = SO_SI.weighted(weights_2D).mean(dim=(lat_name, lon_name))
        ds = ds.drop_vars(["thetao", "so"])
        return ds



import numpy as np
import pytest
import xarray as xr

from ilamb3.analysis import nbp_analysis
from ilamb3.exceptions import MissingVariable, TemporalOverlapIssue
from ilamb3.tests.test_compare import generate_test_dset


def test_nbp():
    grid = dict(ntime=36)
    ref = generate_test_dset(**grid).rename_vars({"da": "nbp"})
    uncert = generate_test_dset(seed=2, **grid)["da"] * 1e-2
    ref["nbp_bnds"] = xr.DataArray(
        np.array(
            [ref["nbp"].values - uncert.values, ref["nbp"].values + uncert.values]
        ).T,
        coords={"time": ref["time"]},
        dims=["time", "nb"],
    )
    ref["nbp"].attrs["bounds"] = "nbp_bnds"
    com = generate_test_dset(seed=3, **grid).rename_vars({"da": "nbp"})
    analysis = nbp_analysis()
    df, _, _ = analysis(ref, com)
    df = df[df["type"] == "score"]
    assert len(df) == 2
    assert np.allclose(
        df[df["name"] == "Difference Score"].iloc[0]["value"], 2.7143968145915746e-08
    )
    assert np.allclose(
        df[df["name"] == "Trajectory Score"].iloc[0]["value"], 4.481852371926836e-08
    )


def test_fail():
    grid = dict(ntime=36)
    ref = generate_test_dset(**grid).rename_vars({"da": "nbp"})
    uncert = generate_test_dset(seed=2, **grid)["da"] * 1e-2
    ref["nbp_bnds"] = xr.DataArray(
        np.array(
            [ref["nbp"].values - uncert.values, ref["nbp"].values + uncert.values]
        ).T,
        coords={"time": ref["time"]},
        dims=["time", "nb"],
    )
    ref["nbp"].attrs["bounds"] = "nbp_bnds"
    com = (
        generate_test_dset(seed=3, **grid)
        .rename_vars({"da": "nbp"})
        .isel(time=slice(24, 36))
    )
    analysis = nbp_analysis()
    with pytest.raises(TemporalOverlapIssue):
        analysis(ref, com)


def test_names():
    grid = dict(ntime=36)
    ref = generate_test_dset(**grid).rename_vars({"da": "nbp"})
    uncert = generate_test_dset(seed=2, **grid)["da"] * 1e-2
    ref["nbp_bnds"] = xr.DataArray(
        np.array(
            [ref["nbp"].values - uncert.values, ref["nbp"].values + uncert.values]
        ).T,
        coords={"time": ref["time"]},
        dims=["time", "nb"],
    )
    ref["nbp"].attrs["bounds"] = "nbp_bnds"
    com = generate_test_dset(seed=3, **grid)
    analysis = nbp_analysis()
    analysis(ref, com.rename_vars({"da": "netAtmosLandCO2Flux"}))
    with pytest.raises(MissingVariable):
        analysis(ref, com.rename_vars({"da": "blah"}))



"""
Runoff sensitivity to temperature and precipitation per river basin.

See Also
--------
ILAMBAnalysis : The abstract base class from which this derives.
"""

from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
import statsmodels.formula.api as smf
import xarray as xr

import ilamb3
import ilamb3.dataset as dset
from ilamb3.analysis.base import ILAMBAnalysis
from ilamb3.exceptions import MissingRegion, MissingVariable
from ilamb3.regions import Regions


def compute_runoff_sensitivity(
    ds: xr.Dataset,
    basins: list[str],
    window_size: int = 9,
    use_cross: bool = True,
    quiet: bool = False,
) -> pd.DataFrame:
    """
    Compute the runoff sensitivites in the basins provided.

    Parameters
    ----------
    ds : xr.Dataset
        The dataset containing `mrro` (runoff), `pr` (precip), and `tas` (temperature).
    basins : list[str]
        The basins in which to compute sensitivities. These should refer to labels that
        have been registered in the ilamb3.regions.Regions class.
    window_size : int
        The number of water years to use in the windowed average.
    use_cross : bool
        Enable to use the cross term `1 + pr + tas + pr * tas` sensitivities, otherwise
        use those from a `1 + pr + tas` least squares fit.
    quiet : bool, default False
        Enable to suppress the progress bar.

    Returns
    -------
    pd.DataFrame
        A dataframe with the sensitivities over each basin.
    """
    # Check that the required variables are present
    required_vars = ["mrro", "tas", "pr"]
    missing = set(required_vars) - set(ds)
    if missing:
        raise MissingVariable(f"Input dataset is lacking variables: {missing}")

    # Check that the basins are in fact registed in the ilamb system
    ilamb_regions = Regions()
    missing = set(basins) - set(ilamb_regions.regions)
    if missing:
        raise MissingRegion(f"Input basins are not registered as regions: {missing}")

    # Associate cell/time measures with the dataset if not present
    ds = dset.shift_lon(ds)
    space = [dset.get_dim_name(ds, "lat"), dset.get_dim_name(ds, "lon")]
    if "cell_measures" not in ds:
        ds["cell_measures"] = dset.compute_cell_measures(ds).pint.dequantify()
    if "time_measures" not in ds:
        ds["time_measures"] = dset.compute_time_measures(ds).pint.dequantify()
        if "bounds" in ds[dset.get_dim_name(ds, "time")].attrs:
            bnds = ds[dset.get_dim_name(ds, "time")].attrs["bounds"]
            if bnds in ds:
                ds = ds.drop_vars(bnds)

    # Compute the water year for the time series (beginning of year is Oct)
    t = ds[dset.get_dim_name(ds, "time")]
    ds["water_year"] = xr.where(t.dt.month > 9, t.dt.year, t.dt.year - 1)
    ds = ds.set_coords("water_year")

    # Averages per water year. This is weighted by the number of days per month in each
    # water year. The `groupby` does not work with the `weighted` accessor and so we
    # take the weighted mean by hand.
    ds = (
        (ds * ds["time_measures"]).groupby("water_year").sum()
        / ds["time_measures"].groupby("water_year").sum()
    ).drop_vars("time_measures")

    # Loop over basins and build up the dataframe of sensitivities
    df = []
    for basin in basins:
        # Compute the regional mean values per basin
        dsb = ilamb_regions.restrict_to_region(ds, basin)
        msr = dsb["cell_measures"].fillna(0)
        dsb = dsb.drop_vars([v for v in ds.data_vars if v not in required_vars])
        dsb = dsb.weighted(msr).mean(dim=space)

        # Take a windowed (decadal) average over the water years
        mean = dsb.rolling(water_year=window_size, center=True).mean()

        # Compute the anomalies
        anomaly = mean - mean.mean()
        anomaly["mrro"] = anomaly["mrro"] / mean["mrro"].mean() * 100.0
        anomaly["pr"] = anomaly["pr"] / mean["pr"].mean() * 100.0

        # Fit a linear model (with and without the cross term) and compute stats
        model = smf.ols("mrro ~ tas + pr", data=anomaly.to_dataframe()).fit()
        model_cross = smf.ols("mrro ~ tas * pr", data=anomaly.to_dataframe()).fit()
        focus = model_cross if use_cross else model
        out = {
            "basin": basin,
            "tas Sensitivity": focus.params.to_dict()["tas"],
            "tas Low": focus.conf_int()[0].to_dict()["tas"],
            "tas High": focus.conf_int()[1].to_dict()["tas"],
            "pr Sensitivity": focus.params.to_dict()["pr"],
            "pr Low": focus.conf_int()[0].to_dict()["pr"],
            "pr High": focus.conf_int()[1].to_dict()["pr"],
            "R2": model.rsquared,
            "R2 Cross": model_cross.rsquared,
            "Cond": focus.condition_number,
        }
        df.append(out)
    df = pd.DataFrame(df).set_index("basin")
    return df


class runoff_sensitivity_analysis(ILAMBAnalysis):
    """
    Runoff sensitivity to temperature and precipitation per river basin.

    Parameters
    ----------
    basin_source : str
        The source file for the basins to use in the analysis.
    sensitivity_frame : pd.DataFrame, optional
        The reference sensitivities. If not provided, we will compute them from the
        sources provided.
    mrro_source : str
        The source of runoff (mrro), required if no sensitivities are provided.
    tas_source : str
        The source of temperature (tas), required if no sensitivities are provided.
    pr_source : str
        The source of precipitation (pr), required if no sensitivities are provided.
    timestamp_start : str
        The initial time over which we compute comparison sensitivities.
    timestamp_end : str
        The final time over which we compute comparison sensitivities.

    Methods
    -------
    required_variables
        What variables are required.
    __call__
        The method
    """

    def __init__(
        self,
        basin_source: str | Path,
        sensitivity_frame: None | str | pd.DataFrame = None,
        mrro_source: str | Path = None,
        tas_source: str | Path = None,
        pr_source: str | Path = None,
        timestamp_start: str = "1940-10-01",
        timestamp_end: str = "2014-10-01",
    ):  # numpydoc ignore=GL08
        # Consistency checks
        cat = ilamb3.ilamb_catalog()
        self.sources = []
        if sensitivity_frame is None:
            if [src for src in [mrro_source, tas_source, pr_source] if src is None]:
                raise ValueError(
                    f"If sensitivities are not provided, you must give a source for each variable. Found {mrro_source=} {tas_source=} {pr_source=}"
                )
            for src in [mrro_source, tas_source, pr_source]:
                if Path(src).is_file():
                    self.sources.append(xr.open_dataset(src))
                elif src in cat:
                    self.sources.append(cat[src].read())
                else:
                    raise ValueError(f"Could not load source file {src=}")

        # Load/store the sensitivities if present
        self.sensitivity_frame = sensitivity_frame
        if sensitivity_frame is not None:
            read = Path(sensitivity_frame).suffix.replace(".", "read_")
            if read not in dir(pd):
                raise OSError(f"Cannot read sensitivty file: {sensitivity_frame}")
            read = pd.__dict__[read]
            self.sensitivity_frame = read(sensitivity_frame)

        # Register basins in the ILAMB region system
        ilamb_regions = Regions()
        if Path(basin_source).is_file():
            self.basins = ilamb_regions.add_netcdf(basin_source)
        elif isinstance(basin_source, str) and basin_source in cat:
            self.basins = ilamb_regions(cat[basin_source].read())
        else:
            raise ValueError(f"Unable to ingest {basin_source=} as ILAMB regions.")

        # Record timestamp variables
        self.timestamp_start = timestamp_start
        self.timestamp_end = timestamp_end

    def required_variables(self) -> list[str]:
        """
        Return the variable names required in this analysis.

        Returns
        -------
        list
            A list of the required variables, here always [pr, tas, mrro].
        """
        return ["pr", "tas", "mrro"]

    def __call__(
        self, ref: xr.Dataset, com: xr.Dataset, basins: list[str] | None = None
    ) -> tuple[pd.DataFrame, xr.Dataset, xr.Dataset]:
        """
        Apply the ILAMB bias methodology on the given datasets.

        Parameters
        ----------
        ref : xr.Dataset
            Not used in this analysis object.
        com : xr.Dataset
            The comparison dataset.
        basins : list[str], optional
            The basins in which to compute sensitivities. By default will use all those
            present in the provided `basin_source` but a subset can be provided here if
            desired.

        Returns
        -------
        pd.DataFrame
            A dataframe with scalar and score information from the comparison.
        xr.Dataset
            A dataset containing reference grided information from the comparison.
        xr.Dataset
            A dataset containing comparison grided information from the comparison.
        """
        # Ensure we have the variables we need for this confrontation
        missing = set(self.required_variables()) - set(com)
        if missing:
            raise MissingVariable(f"Comparison dataset is lacking variables: {missing}")
        basins = self.basins if basins is None else basins

        # Optionally compute reference sensitivities if not given
        if self.sensitivity_frame is None:
            ref = xr.merge(self.sources)
            self.sensitivity_frame = compute_runoff_sensitivity(ref, basins)
        df_ref = self.sensitivity_frame

        # Compute comparisons sensitivities
        com = com.sel({"time": slice(self.timestamp_start, self.timestamp_end)})
        df_com = compute_runoff_sensitivity(com, basins)

        # Paint by the numbers to create maps of scalars
        ilamb_regions = Regions()

        def _df_to_ds(df):  # numpydoc ignore=GL08
            ds = xr.Dataset(
                {
                    key: ilamb_regions.region_scalars_to_map(df[key].to_dict())
                    for key in df.columns
                }
            )
            return ds

        ds_ref = _df_to_ds(df_ref)
        ds_com = _df_to_ds(df_com)

        return df_com, ds_ref, ds_com

    def plots(
        self, ds_ref: xr.Dataset, dsd_com: dict[str, xr.Dataset]
    ) -> dict[str, dict[str, plt.Figure]]:
        """
        Return figures of the reference and comparison data.

        Parameters
        ----------
        ds_ref : xr.Dataset
            The reference dataset.
        dsd_com : dictionary of xr.Dataset
            A dictionary of xr.Datasets whose keys are the comparisons (models).

        Returns
        -------
        dict[str,dict[str,plt.Figure]]
            A nested dictionary of matplotlib figures where the first level corresponds
            to the source name and the second level the plot name.
        """
        pass


if __name__ == "__main__":
    from ilamb3.models import ModelESGF

    # Initialize the analysis
    ref_file = Path("ref.parquet")
    analysis = runoff_sensitivity_analysis(
        basin_source="mrb_basins.nc",
        sensitivity_frame=ref_file if ref_file.is_file() else None,
        mrro_source="mrro | LORA",
        tas_source="tas | CRU4.02",
        pr_source="pr | GPCPv2.3",
    )

    # Test against CanESM5
    m = ModelESGF("CanESM5", "r1i1p1f1", "gn")
    com = xr.merge(
        [m.get_variable(v) for v in analysis.required_variables()], compat="override"
    )

    # Run the analysis
    df, ds_ref, ds_com = analysis(xr.Dataset(), com)
    if not ref_file.is_file():
        analysis.sensitivity_frame.to_parquet(ref_file)

    # Make plots
    figs = analysis.plots(ds_ref, {"CanESM5": ds_com})

    # Hanjun's numbers for reference:
    #   tas: -0.61410
    #   pr: 1.233



from typing import Any

import numpy as np
import pandas as pd
import xarray as xr

import ilamb3.compare as cmp
import ilamb3.dataset as dset
import ilamb3.plot as ilplt
from ilamb3.analysis.base import ILAMBAnalysis


class accumulate_analysis(ILAMBAnalysis):
    def __init__(self, required_variable: str, **kwargs: Any):
        self.required_variable = required_variable
        self.description = kwargs.get("description", "Period Mean")
        self.kwargs = kwargs

    def required_variables(self):
        return [self.required_variable]

    def __call__(self, ref: xr.Dataset, com: xr.Dataset):
        varname = self.required_variable
        unit = ref[varname].attrs["units"]

        # make time series comparable
        ref, com = cmp.trim_time(ref, com)
        com = dset.convert(com, unit, varname=varname)

        # find period mean
        ref_mean = dset.integrate_time(ref, varname, mean=True)
        com_mean = dset.integrate_time(com, varname, mean=True)

        # score the difference as a bias
        bias = com_mean - ref_mean
        bias_score = np.exp(-np.abs((bias) / ref_mean))

        # create a dataframe of scalars
        df = pd.DataFrame(
            [
                {
                    "source": "Reference",
                    "region": "None",
                    "analysis": "accumulate",
                    "name": self.description,
                    "type": "scalar",
                    "units": unit,
                    "value": float(ref_mean),
                },
                {
                    "source": "Comparison",
                    "region": "None",
                    "analysis": "accumulate",
                    "name": self.description,
                    "type": "scalar",
                    "units": unit,
                    "value": float(com_mean),
                },
                {
                    "source": "Comparison",
                    "region": "None",
                    "analysis": "accumulate",
                    "name": "Bias",
                    "type": "scalar",
                    "units": unit,
                    "value": float(bias),
                },
                {
                    "source": "Comparison",
                    "region": "None",
                    "analysis": "accumulate",
                    "name": "Bias Score",
                    "type": "score",
                    "units": "1",
                    "value": float(bias_score),
                },
            ]
        )

        # When writing intermediate outputs, make sure dims are a consistent name
        ref, com = cmp.rename_dims(ref, com)

        return df, ref, com

    def plots(
        self,
        df: pd.DataFrame,
        ref: xr.Dataset,
        com: dict[str, xr.Dataset],
    ) -> pd.DataFrame:
        # Setup plot data
        varname = self.required_variable
        com["Reference"] = ref
        lim = ilplt.determine_plot_limits(com, percent_pad=0).set_index("name")
        lim.loc[varname, ["cmap", "title"]] = [None, self.description]

        # Build up a dataframe of matplotlib axes
        axs = [
            {
                "name": plot,
                "title": lim.loc[plot, "title"],
                "region": None,
                "source": source,
                "axis": (
                    ilplt.plot_curve(
                        {source: ds} | {"Reference": ref},
                        plot,
                        vmin=lim.loc[plot, "low"]
                        - 0.05 * (lim.loc[plot, "high"] - lim.loc[plot, "low"]),
                        vmax=lim.loc[plot, "high"]
                        + 0.05 * (lim.loc[plot, "high"] - lim.loc[plot, "low"]),
                        title=source + " - " + lim.loc[plot, "title"],
                    )
                    if plot in ds
                    else pd.NA
                ),
            }
            for plot in [varname]
            for source, ds in com.items()
            if source != "Reference"
        ]

        # Plot all curves on one
        axs += [
            {
                "name": f"all{varname}",
                "title": self.description,
                "region": None,
                "source": None,
                "axis": (
                    ilplt.plot_curve(
                        com,
                        varname,
                        vmin=lim.loc[varname, "low"]
                        - 0.05 * (lim.loc[varname, "high"] - lim.loc[varname, "low"]),
                        vmax=lim.loc[varname, "high"]
                        + 0.05 * (lim.loc[varname, "high"] - lim.loc[varname, "low"]),
                        title=f"Combined {self.description}",
                    )
                ),
            }
        ]
        axs = pd.DataFrame(axs).dropna(subset=["axis"])
        return axs



"""
An ILAMB transform to extract AMOC out of msftmz for comparison to the RAPID data.
"""

import xarray as xr

import ilamb3.dataset as dset
from ilamb3.transform.base import ILAMBTransform


class msftmz_to_rapid(ILAMBTransform):
    """
    An ILAMB transform to extract AMOC out of msftmz for comparison to the RAPID data.
    """

    def __init__(self):
        pass

    def required_variables(self) -> list[str]:
        """
        Return the variables this transform uses.
        """
        return ["msftmz"]

    def __call__(self, ds: xr.Dataset) -> xr.Dataset:
        """
        Extract AMOC strength for comparison to RAPID.
        """
        if "amoc" in ds:
            return ds
        if "msftmz" not in ds:
            return ds
        lat_name = dset.get_dim_name(ds, "lat")
        depth_name = dset.get_dim_name(ds, "depth")
        ds["amoc"] = dset.convert(
            ds["msftmz"]
            .isel(basin=0)
            .sel({lat_name: 26.5}, method="nearest")
            .max(depth_name),
            "Sv",
            "msftmz",
        )
        ds = ds.drop_vars("msftmz")
        return ds



import numpy as np
import pandas as pd
import xarray as xr

import ilamb3.compare as cmp


def generate_test_dset(seed: int = 1, ntime=None, nlat=None, nlon=None):
    rs = np.random.RandomState(seed)
    coords = []
    dims = []
    if ntime is not None:
        time = pd.date_range(start="2000-01-15", periods=ntime, freq="30D")
        coords.append(time)
        dims.append("time")
    if nlat is not None:
        lat = np.linspace(-90, 90, nlat + 1)
        lat = 0.5 * (lat[1:] + lat[:-1])
        coords.append(lat)
        dims.append("lat")
    if nlon is not None:
        lon = np.linspace(-180, 180, nlon + 1)
        lon = 0.5 * (lon[1:] + lon[:-1])
        coords.append(lon)
        dims.append("lon")
    ds = xr.Dataset(
        data_vars={
            "da": xr.DataArray(
                rs.rand(*[len(c) for c in coords]) * 1e-8, coords=coords, dims=dims
            ),
        }
    )
    ds["da"].attrs["units"] = "kg m-2 s-1"
    return ds


def test_nest_spatial_grids():
    ds1 = generate_test_dset(nlat=2, nlon=3)
    ds2 = generate_test_dset(nlat=5, nlon=7)
    ds1 = ds1.cf.add_bounds("lat")
    ds1_, ds2_ = cmp.nest_spatial_grids(ds1, ds2)
    assert np.allclose((ds1_ - ds2_).sum()["da"], -6.06395917e-08)
    dsa, dsb = cmp.pick_grid_aligned(ds1, ds2, ds1_, ds2_)
    xr.testing.assert_allclose(dsa, ds1_)
    xr.testing.assert_allclose(dsb, ds2_)


def test_spatial_align():
    ds1 = generate_test_dset(nlat=2, nlon=3)
    ds2 = generate_test_dset(nlat=3, nlon=4)
    assert not cmp.is_spatially_aligned(ds1, ds2)
    ds2 = generate_test_dset(nlat=2, nlon=4)
    assert not cmp.is_spatially_aligned(ds1, ds2)
    ds2 = generate_test_dset(nlat=2, nlon=3)
    ds2["lat"] = [1, 2]
    assert not cmp.is_spatially_aligned(ds1, ds2)
    ds2 = generate_test_dset(nlat=2, nlon=3)
    ds2["lon"] = [1, 2, 3]


def test_grid_align():
    ds1 = generate_test_dset(nlat=2, nlon=3)
    ds2 = generate_test_dset(nlat=2, nlon=3)
    ds1_, ds2_ = cmp.pick_grid_aligned(ds1, ds2)
    xr.testing.assert_allclose(ds1_, ds2_)
    ds2 = generate_test_dset(nlat=3, nlon=4)
    ds1_, ds2_ = cmp.pick_grid_aligned(ds1, ds2)
    xr.testing.assert_allclose(ds1_["lat"], ds2_["lat"])
    xr.testing.assert_allclose(ds1_["lon"], ds2_["lon"])


def test_trim_time():
    ds = {"1": generate_test_dset(ntime=2), "2": generate_test_dset(ntime=3)}
    out = cmp.trim_time(**ds)
    assert len(out) == 2
    assert all([ds["time"].size == 2 for _, ds in out.items()])
    out1, out2 = cmp.trim_time(generate_test_dset(ntime=4), **ds)
    assert len(out1) == 1
    assert len(out2) == 2


def test_same_spatial_grid():
    grid = generate_test_dset(nlat=2, nlon=2)
    out = cmp.same_spatial_grid(
        grid,
        generate_test_dset(nlat=2, nlon=3),
    )
    xr.testing.assert_allclose(grid["lat"], out[0]["lat"])
    xr.testing.assert_allclose(grid["lon"], out[0]["lon"])
    out = cmp.same_spatial_grid(
        grid,
        **{"1": generate_test_dset(nlat=3, nlon=2)},
    )
    xr.testing.assert_allclose(grid["lat"], out["1"]["lat"])
    xr.testing.assert_allclose(grid["lon"], out["1"]["lon"])
    out, outkw = cmp.same_spatial_grid(
        grid,
        generate_test_dset(nlat=2, nlon=3),
        **{"1": generate_test_dset(nlat=3, nlon=2)},
    )
    xr.testing.assert_allclose(grid["lat"], out[0]["lat"])
    xr.testing.assert_allclose(grid["lon"], out[0]["lon"])
    xr.testing.assert_allclose(grid["lat"], outkw["1"]["lat"])
    xr.testing.assert_allclose(grid["lon"], outkw["1"]["lon"])



"""Modular ILAMB methodology functions."""

from ilamb3.analysis.accumulate import accumulate_analysis
from ilamb3.analysis.base import add_overall_score
from ilamb3.analysis.bias import bias_analysis
from ilamb3.analysis.cycle import cycle_analysis
from ilamb3.analysis.hydro import hydro_analysis
from ilamb3.analysis.nbp import nbp_analysis
from ilamb3.analysis.relationship import relationship_analysis
from ilamb3.analysis.rmse import rmse_analysis
from ilamb3.analysis.runoff_sensitivity import runoff_sensitivity_analysis
from ilamb3.analysis.spatial_distribution import spatial_distribution_analysis
from ilamb3.analysis.timeseries import timeseries_analysis

DEFAULT_ANALYSES = {
    "Bias": bias_analysis,
    "RMSE": rmse_analysis,
    "Annual Cycle": cycle_analysis,
    "Spatial Distribution": spatial_distribution_analysis,
}
ALL_ANALYSES = {
    "nbp": nbp_analysis,
    "Runoff Sensitivity": runoff_sensitivity_analysis,
    "Hydro": hydro_analysis,
    "timeseries": timeseries_analysis,
    "accumulate": accumulate_analysis,
    "relationship": relationship_analysis,
} | DEFAULT_ANALYSES

__all__ = list(ALL_ANALYSES.keys()) + ["add_overall_score"]



from itertools import chain
from pathlib import Path
from typing import Any

import matplotlib.pyplot as mpl
import numpy as np
import pandas as pd
import xarray as xr
from loguru import logger

import ilamb3.compare as cmp
import ilamb3.dataset as dset
import ilamb3.plot as plt
from ilamb3.analysis.base import ILAMBAnalysis, scalarify


def metric_maps(
    da: xr.Dataset | xr.DataArray, varname: str | None = None
) -> xr.Dataset:
    """
    Return a dataset containing Deeksha's metrics request for the ILAMB Hydro project.

    Parameters
    ----------
    da : xr.Dataset or xr.DataArray
        The dataset containing the variable
    varname: str, optional
        The name of the variable if a xr.Dataset is passed.

    Returns
    -------
    xr.Dataset
        The metrics derived from the input dataset.

    """
    if isinstance(da, xr.Dataset):
        assert varname is not None
        da = da[varname]
    out = {}

    # annual
    grp = da.groupby("time.year")
    out["annual_mean"] = grp.mean().mean(dim="year")
    out["annual_std"] = grp.mean().std(dim="year")

    # seasons
    grp = da.groupby("time.season")
    mean = grp.mean()
    out.update(
        {
            f"seasonal_mean_{str(s)}": mean.sel(season=s).drop_vars("season")
            for s in mean["season"].values
        }
    )
    std = grp.std("time")
    out.update(
        {
            f"seasonal_std_{str(s)}": std.sel(season=s).drop_vars("season")
            for s in std["season"].values
        }
    )
    return xr.Dataset(out)


def score_difference(ref: xr.Dataset, com: xr.Dataset) -> xr.Dataset:
    # Compute differences and scores
    ref_, com_ = cmp.nest_spatial_grids(ref, com)
    diff = com_ - ref_
    diff = diff.rename_vars({v: f"{v}_difference" for v in diff})
    # Add scores to the means that also have std's
    diff = diff.merge(
        {
            v.replace("_difference", "_score"): np.exp(
                -np.abs(diff[v])
                / ref_[v.replace("_mean", "_std").replace("_difference", "")]
            )
            for v in diff
            if "mean" in v and v.replace("_mean_", "_std_") in diff
        }
    )
    # Set the units of these scores
    for var, da in diff.items():
        if "score" not in var:
            continue
        diff[var].attrs["units"] = "1"
    # Rename the lat dimension for merging with the comparison on return
    lat_name = dset.get_dim_name(diff, "lat")
    lon_name = dset.get_dim_name(diff, "lon")
    com = com.merge(diff.rename({lat_name: f"{lat_name}_", lon_name: f"{lon_name}_"}))
    return com


def generate_titles(qname: str) -> str:
    """
    Transform the quantity name into a display string.
    """
    tokens = [v.capitalize() if v.islower() else v for v in qname.split("_")]
    if (
        len(tokens) > 2
        and tokens[0] == "Seasonal"
        and tokens[2] in ["DJF", "MAM", "JJA", "SON"]
    ):
        tokens[0] = {"DJF": "Winter", "MAM": "Spring", "JJA": "Summer", "SON": "Fall"}[
            tokens.pop(2)
        ]
    title = " ".join(tokens)
    print(f"{qname:30} | {title}")
    return title


class hydro_analysis(ILAMBAnalysis):
    def __init__(
        self,
        required_variable: str,
        regions: list[str] | None = None,
        output_path: Path | None = None,
        **kwargs: Any,
    ):
        self.req_variable = required_variable
        self.regions = regions if isinstance(regions, list) else [None]
        self.output_path = output_path

        # This analysis will split plots/scalars into sections as organized below
        self.sections = {
            "Annual": [f"mean_{region}" for region in self.regions]
            + [
                "annual_mean",
                "annual_mean_difference",
                "annual_mean_score",
                "annual_std",
                "annual_std_difference",
            ],
            "Winter (DJF)": [
                "seasonal_mean_DJF",
                "seasonal_mean_DJF_difference",
                "seasonal_mean_DJF_score",
                "seasonal_std_DJF",
                "seasonal_std_DJF_difference",
            ],
            "Spring (MAM)": [
                "seasonal_mean_MAM",
                "seasonal_mean_MAM_difference",
                "seasonal_mean_MAM_score",
                "seasonal_std_MAM",
                "seasonal_std_MAM_difference",
            ],
            "Summer (JJA)": [
                "seasonal_mean_JJA",
                "seasonal_mean_JJA_difference",
                "seasonal_mean_JJA_score",
                "seasonal_std_JJA",
                "seasonal_std_JJA_difference",
            ],
            "Fall (SON)": [
                "seasonal_mean_SON",
                "seasonal_mean_SON_difference",
                "seasonal_mean_SON_score",
                "seasonal_std_SON",
                "seasonal_std_SON_difference",
            ],
        }

    def required_variables(self) -> list[str]:
        """
        Return the list of variables required for this analysis.

        Returns
        -------
        list
            The variable names used in this analysis.
        """
        return [self.req_variable]

    def _get_analysis_section(self, varname: str) -> str:
        """Given the plot/variable, return from which section it belongs."""
        section = [s for s, vs in self.sections.items() if varname in vs]
        if not section:
            raise ValueError(f"Could not find {varname} in {self.sections}.")
        return section[0]

    def _make_comparable(
        self, ref: xr.Dataset, com: xr.Dataset
    ) -> tuple[xr.Dataset, xr.Dataset]:
        if dset.is_temporal(ref):
            ref, com = cmp.trim_time(ref, com)
        # ensure longitudes are uniform
        ref, com = cmp.adjust_lon(ref, com)
        # ensure the lat/lon dims are sorted
        if dset.is_spatial(ref):
            ref = ref.sortby(
                [dset.get_dim_name(ref, "lat"), dset.get_dim_name(ref, "lon")]
            )
        if dset.is_spatial(com):
            com = com.sortby(
                [dset.get_dim_name(com, "lat"), dset.get_dim_name(com, "lon")]
            )
        # convert units
        com = dset.convert(
            com, ref[self.req_variable].attrs["units"], varname=self.req_variable
        )
        return ref, com

    def __call__(
        self,
        ref: xr.Dataset,
        com: xr.Dataset,
    ) -> tuple[pd.DataFrame, xr.Dataset, xr.Dataset]:
        """
        Apply the ILAMB bias methodology on the given datasets.
        """
        # Initialize
        varname = self.req_variable

        # Make the variables comparable
        ref_, com_ = self._make_comparable(ref, com)

        # Run the hydro metrics, read cached reference if running in batch mode
        if (
            self.output_path is not None
            and (self.output_path / "Reference.nc").is_file()
        ):
            logger.info(
                f"Reading in cached reference data: {self.output_path / 'Reference.nc'}"
            )
            ref = xr.open_dataset(self.output_path / "Reference.nc")
        else:
            ref = metric_maps(ref_, varname)
        com = metric_maps(com_, varname)

        # Ensure that arrays are now in memory and score
        logger.info("Computing hydro metrics...")
        ref.load()
        com.load()
        com = score_difference(ref, com)

        # Create scalars
        df = []
        for source, ds in {"Reference": ref, "Comparison": com}.items():
            for vname, da in ds.items():
                if vname.startswith("mean_"):
                    continue
                for region in self.regions:
                    scalar, unit = scalarify(da, vname, region=region, mean=True)
                    df.append(
                        [
                            source,
                            str(region),
                            self._get_analysis_section(vname),
                            generate_titles(vname),
                            "score" if "score" in vname else "scalar",
                            unit,
                            scalar,
                        ]
                    )

        # Compute the regional means
        for region in self.regions:
            if f"mean_{region}" not in ref:
                ref[f"mean_{region}"] = dset.compute_monthly_mean(
                    dset.integrate_space(
                        ref_,
                        varname,
                        region=region,
                        mean=True,
                    )
                )
            com[f"mean_{region}"] = dset.compute_monthly_mean(
                dset.integrate_space(
                    com_,
                    varname,
                    region=region,
                    mean=True,
                )
            )

        # Convert to dataframe
        df = pd.DataFrame(
            df,
            columns=[
                "source",
                "region",
                "analysis",
                "name",
                "type",
                "units",
                "value",
            ],
        )
        df.attrs = self.__dict__.copy()
        return df, ref, com

    def plots(
        self,
        df: pd.DataFrame,
        ref: xr.Dataset,
        com: dict[str, xr.Dataset],
    ) -> pd.DataFrame:
        com["Reference"] = ref

        def _choose_cmap(plot_name):
            if "score" in plot_name:
                return "plasma"
            if "difference" in plot_name:
                return "bwr"
            return "viridis"

        # Which plots are we handling in here? I am building this list from a
        # section layout I created in the constructor.
        plots = list(chain(*[vs for _, vs in self.sections.items()]))

        # Setup plots
        df = plt.determine_plot_limits(com, symmetrize=["difference"]).set_index("name")
        df["title"] = [generate_titles(plot) for plot in df.index]
        df["cmap"] = df.index.map(_choose_cmap)

        # Plot the maps, saving if requested on the fly
        axs = []
        logger.info("Plotting maps...")
        for plot in plots:
            for source, ds in com.items():
                if plot not in ds:
                    continue
                if not dset.is_spatial(ds[plot]):
                    continue
                for region in self.regions:
                    row = {
                        "name": plot,
                        "title": df.loc[plot, "title"],
                        "region": region,
                        "source": source,
                        "analysis": self._get_analysis_section(plot),
                    }
                    ax = plt.plot_map(
                        ds[plot],
                        region=region,
                        vmin=df.loc[plot, "low"],
                        vmax=df.loc[plot, "high"],
                        cmap=df.loc[plot, "cmap"],
                        title=source + " " + df.loc[plot, "title"],
                    )
                    if self.output_path is None:
                        row["axis"] = ax
                    else:
                        row["axis"] = False
                        fig = ax.get_figure()
                        fig.savefig(
                            self.output_path
                            / f"{row['source']}_{row['region']}_{row['name']}.png"
                        )
                        mpl.close(fig)
                    axs.append(row)

        # Plot the curves, saving if requested on the fly
        logger.info("Plotting curves...")
        for plot in [f"mean_{region}" for region in self.regions]:
            region = plot.split("_")[-1]
            for source, ds in com.items():
                if source == "Reference":
                    continue
                if plot not in ds:
                    continue
                if not dset.is_temporal(ds[plot]):
                    continue
                row = {
                    "name": "mean",
                    "title": "Regional Mean",
                    "region": region,
                    "source": source,
                    "analysis": "Annual",
                }
                # Try to set the ylabel but if it fails, use xarray default
                ylabel = None
                try:
                    ylabel = (
                        f"{ref[plot].attrs['long_name']} {ref[plot].attrs['units']}"
                    )
                except Exception:
                    pass
                ax = plt.plot_curve(
                    {source: ds} | {"Reference": ref},
                    plot,
                    vmin=df.loc[plot, "low"]
                    - 0.05 * (df.loc[plot, "high"] - df.loc[plot, "low"]),
                    vmax=df.loc[plot, "high"]
                    + 0.05 * (df.loc[plot, "high"] - df.loc[plot, "low"]),
                    title=f"{source} Regional Mean",
                    ylabel=ylabel,
                )
                if self.output_path is None:
                    row["axis"] = ax
                    continue
                else:
                    row["axis"] = False
                    fig = ax.get_figure()
                    fig.savefig(
                        self.output_path
                        / f"{row['source']}_{row['region']}_{row['name']}.png"
                    )
                    mpl.close(fig)
                axs.append(row)
        axs = pd.DataFrame(axs).dropna(subset=["axis"])
        return axs



"""Functions for rendering ilamb3 output."""

import importlib
import re
from itertools import chain
from pathlib import Path
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pooch
import xarray as xr
import yaml
from jinja2 import Template
from loguru import logger

import ilamb3
import ilamb3.analysis as anl
import ilamb3.compare as cmp
import ilamb3.dataset as dset
import ilamb3.regions as ilr
from ilamb3.analysis.base import ILAMBAnalysis, add_overall_score
from ilamb3.exceptions import AnalysisNotAppropriate, VarNotInModel
from ilamb3.transform import ALL_TRANSFORMS
from ilamb3.transform.base import ILAMBTransform


def fix_pint_units(ds: xr.Dataset) -> xr.Dataset:
    def _fix(units: str) -> str:
        """
        Modify units that pint cannot handle.
        """
        try:
            val_units = float(units)
        except ValueError:
            return units
        if np.allclose(val_units, 1):
            return "dimensionless"
        if np.allclose(val_units, 1e-3):
            return "psu"
        return units

    for var, da in ds.items():
        if "units" not in da.attrs:
            continue
        ds[var].attrs["units"] = _fix(da.attrs["units"])
    return ds


def select_analysis_variable(setup: dict[str, Any]) -> str:
    """
    Return the main variable to be used in this analysis.

    We will try to guess it from what the user has specified, but to avoid
    ambiguity please specify `analysis_variable` in your configure block.

    Parameters
    ----------
    setup: dict
        A dictionary of keywords parsed from the configure files.

    Returns
    -------
    str
        The main setup variable.

    """
    variable = setup.get("analysis_variable", None)
    if variable is None and "sources" in setup:
        if len(setup["sources"]) > 1:
            raise ValueError(
                "Ambiguous definition of the main variable of this analysis. "
                "If you have multiple sources, you should specify `analysis_variable` "
                f"in your configure file. This is the problematic portion:\n{yaml.dump(setup)}"
            )
        variable = next(iter(setup["sources"]))
    return variable


def setup_analyses(
    setup: dict[str, Any], output_path: Path | None
) -> dict[ILAMBAnalysis]:
    """
    Return the initialized analysis components to be used for this block.

    You may list individual analyses to run in a `analyses` line in the
    configure block or we will run the ilamb3.analysis.DEFAULT_ANALYSES by
    default. Note that we have special rules for invoking the `relationship`
    analysis. It is run by including a configure line similar to `sources` but
    call `relationships` for each independent variable you wish to compare to
    the main variable of this block.

    Parameters
    ----------
    setup: dict
        A dictionary of keywords parsed from the configure files.

    Returns
    -------
    dict
        A dictionary of initialized ilamb3.analysis.base.ILAMBAnalysis objects.

    """
    main_variable = select_analysis_variable(setup)
    analyses = setup.get("analyses", list(anl.DEFAULT_ANALYSES.keys()))
    invalid = set(analyses) - set(anl.ALL_ANALYSES)
    if invalid:
        raise ValueError(
            f"Invalid analyses given, {invalid} not in {list(anl.ALL_ANALYSES.keys())}. "
            f"This is the problematic portion:\n{yaml.dump(setup)}"
        )
    analyses = {
        a: anl.ALL_ANALYSES[a](
            **(
                setup
                | {
                    "required_variable": main_variable,
                    "output_path": None
                    if ilamb3.conf["run_mode"] == "interactive"
                    else output_path,
                }
            )
        )
        for a in analyses
    }
    if "relationships" in setup:
        analyses.update(
            {
                f"Relationship {ind_variable}": anl.ALL_ANALYSES["relationship"](
                    main_variable, ind_variable, **setup
                )
                for ind_variable in setup["relationships"]
            }
        )
    return analyses


def setup_transforms(setup: dict[str, Any]) -> list[ILAMBTransform]:
    """
    Return the initialized transforms to be used for this block.

    Parameters
    ----------
    setup: dict
        A dictionary of keywords parsed from the configure files.

    Returns
    -------
    list
        A list of initialized ilamb3.transform.base.ILAMBTransform objects.

    """
    transforms = setup.get("transforms", [])
    transform_names = [
        next(iter(t.keys())) if isinstance(t, dict) else t for t in transforms
    ]
    invalid = set(transform_names) - set(ALL_TRANSFORMS)
    if invalid:
        raise ValueError(
            f"Invalid transform{'s' if len(invalid) > 1 else ''} given. "
            f"{list(invalid)} not in {list(ALL_TRANSFORMS.keys())}. "
            f"This is the problematic portion:\n{yaml.dump(setup)}"
        )
    transforms = [
        ALL_TRANSFORMS[t](**args[t]) if isinstance(args, dict) else ALL_TRANSFORMS[t]()
        for t, args in zip(transform_names, transforms)
    ]
    return transforms


def find_related_variables(
    analyses: dict[ILAMBAnalysis],
    transforms: list[ILAMBTransform],
    alternate_vars: list[str] | None,
) -> list[str]:
    """
    Return the list of required variables for the given analyses/transforms.

    Parameters
    ----------
    analyses: dict
        A dictionary of initialized ilamb3.analysis.base.ILAMBAnalysis objects.
    transforms: list
        A list of ilamb3.transform.base.ILAMBTransform to apply.
    alternate_vars: list, optional
        A list of alternate variables that this analysis will accept.

    Returns
    -------
    list
        A list of the related variables to be used in query the model database.

    """
    related = list(
        chain(
            *[a.required_variables() for _, a in analyses.items()],
            *[t.required_variables() for t in transforms],
            [] if alternate_vars is None else alternate_vars,
        )
    )
    related = list(set(related))
    return related


def augment_setup_with_options(
    setup: dict[str, Any], reference_data: pd.DataFrame
) -> dict[str, Any]:
    """
    Augment the configure block with global ilamb3 options.
    """
    # Augment options with things in the global options
    if "regions" not in setup:
        setup["regions"] = ilamb3.conf["regions"]
    if "method" not in setup:
        if ilamb3.conf["prefer_regional_quantiles"]:
            setup["method"] = "RegionalQuantiles"
            try:
                setup["quantile_database"] = pd.read_parquet(
                    reference_data.loc[ilamb3.conf["quantile_database"], "path"]
                )
                setup["quantile_threshold"] = ilamb3.conf["quantile_threshold"]
                ilr.Regions().add_netcdf(
                    xr.load_dataset(reference_data.loc["regions/Whittaker.nc", "path"])
                )
            except Exception:
                setup["method"] = "Collier2018"
        else:
            setup["method"] = "Collier2018"
    if "use_uncertainty" not in setup:
        setup["use_uncertainty"] = ilamb3.conf["use_uncertainty"]
    return setup


def _lookup(df: xr.Dataset, key: str) -> list[str]:
    """
    Lookup the key in the dataframe, allowing that it may be a regular expression.
    """
    try:
        return [df.loc[key, "path"]]
    except KeyError:
        pass
    out = sorted(df[df.index.str.contains(key)]["path"].to_list())
    if not out:
        raise ValueError(f"Could not find {key} in the reference dataframe.")
    return out


def _load_reference_data(
    reference_data: pd.DataFrame,
    variable_id: str,
    sources: dict[str, str],
    relationships: dict[str, str] | None = None,
    transforms: list | None = None,
) -> xr.Dataset:
    """
    Load the reference data into containers and merge if more than 1 variable is
    used.
    """
    # First load all variables defined as `sources` or in `relationships`.
    if relationships is not None:
        sources = sources | relationships
    ref = {
        key: xr.open_mfdataset(_lookup(reference_data, str(filename)))
        for key, filename in sources.items()
    }
    # Sometimes there is a bounds variable but it isn't in the attributes
    ref = {key: dset.fix_missing_bounds_attrs(ds) for key, ds in ref.items()}
    # Merge all the data together
    if len(ref) > 1:
        ref = cmp.trim_time(**ref)
        ref = cmp.same_spatial_grid(ref[variable_id], **ref)
        ds_ref = xr.merge([v for _, v in ref.items()], compat="override")
    else:
        ds_ref = ref[variable_id]
    # pint can't handle some units like `0.001`, so we have to intercept and fix
    ds_ref = fix_pint_units(ds_ref)
    # Finally apply transforms
    for transform in transforms:
        ds_ref = transform(ds_ref)
    if variable_id not in ds_ref:
        raise VarNotInModel(
            f"Could not find or create '{variable_id}' from reference data {list(reference_data['variable_id'].unique())}"
        )
    return ds_ref


def _load_comparison_data(
    df: pd.DataFrame,
    variable_id: str,
    alternate_vars: list[str] | None = None,
    transforms: list | None = None,
) -> xr.Dataset:
    """
    Load the comparison (model) data into containers and merge if more than 1
    variable is used.

    Parameters
    ----------
    df: pd.DataFrame
        The database of all possible variables and where to load them.
    variable_id: str
        The name of the variable that is the focus in the comparison.
    alternate_vars: list[str], optional
        A list of acceptable synonyms to be used if `variable_id` is not found.
    transforms: list, optional
        A list of functions that operate on the combined dataset.
    """
    # First load all variables passed into the input dataframe. This will
    # include all relationship variables as well as alternates.
    com = {
        var: xr.open_mfdataset(sorted((df[df["variable_id"] == var]["path"]).to_list()))
        for var in df["variable_id"].unique()
    }
    # If the variable_id is not present, it may be called something else
    if alternate_vars is not None and variable_id not in com:
        found = [v for v in alternate_vars if v in com]
        if found:
            found = found[0]
            com[variable_id] = (
                com[found].rename_vars({found: variable_id})
                if found in com[found]
                else com[found]
            )
            com.pop(found)
    # Fix bounds attributes (there is a bounds variable but it isn't in the
    # attributes)
    com = {var: dset.fix_missing_bounds_attrs(ds) for var, ds in com.items()}
    # Merge all the data together
    if len(com) > 1:
        # The grids should be the same, but sometimes models generate output
        # with very small differences in lat/lon
        try:
            com = cmp.same_spatial_grid(com[next(iter(com))], **com)
        except KeyError:
            pass
        ds_com = xr.merge([v for _, v in com.items()], compat="override")
    else:
        ds_com = com[next(iter(com))]
    # pint can't handle some units like `0.001`, so we have to intercept and fix
    ds_com = fix_pint_units(ds_com)
    # Finally apply transforms. These may create the needed variable.
    for transform in transforms:
        ds_com = transform(ds_com)
    if variable_id not in ds_com:
        raise VarNotInModel(
            f"Could not find or create '{variable_id}' from model variables {list(df['variable_id'].unique())}"
        )
    return ds_com


def registry_to_dataframe(registry: pooch.Pooch) -> pd.DataFrame:
    """
    Convert a ILAMB/IOMB registry to a DatasetCollection for use in REF.

    Parameters
    ----------
    registry : pooch.Pooch
        The pooch registry.

    Returns
    -------
    DatasetCollection
        The converted collection.
    """
    df = pd.DataFrame(
        [
            {
                "key": key,
                "path": registry.abspath / Path(key),
            }
            for key in registry.registry.keys()
        ]
    )
    return df.set_index("key")


def remove_irrelevant_variables(df: pd.DataFrame, **setup: Any) -> pd.DataFrame:
    """
    Remove unused variables from the dataframe.
    """
    reduce = df[
        df["variable_id"].isin(
            list(setup["sources"].keys())
            + list(setup.get("relationships", {}).keys())
            + setup.get("alternate_vars", [])
            + setup.get("related_vars", [])
        )
    ]
    return reduce


def _load_local_assets(
    csv_file: Path, ref_file: Path, com_file: Path
) -> tuple[pd.DataFrame, xr.Dataset, xr.Dataset]:
    if not (csv_file.is_file() and ref_file.is_file() and com_file.is_file()):
        raise ValueError()
    df = pd.read_csv(str(csv_file))
    df["region"] = df["region"].astype(str).str.replace("nan", "None")
    ds_ref = xr.open_dataset(str(ref_file))
    ds_com = xr.open_dataset(str(com_file))
    return df, ds_ref, ds_com


def run_single_block(
    block_name: str,
    reference_data: pd.DataFrame,
    comparison_data: pd.DataFrame,
    output_path: Path,
    **setup: Any,
):
    """
    Run a configuration block.

    Parameters
    ----------

    """
    # Initialize
    if reference_data.index.name != "key":
        reference_data = reference_data.set_index("key")
    setup = augment_setup_with_options(setup, reference_data)
    variable = select_analysis_variable(setup)
    analyses = setup_analyses(setup, output_path)
    transforms = setup_transforms(setup)

    # Thin out the dataframe to only contain variables we need for this block.
    comparison_data = comparison_data[
        comparison_data["variable_id"].isin(
            find_related_variables(
                analyses, transforms, setup.get("alternate_vars", [])
            )
        )
    ]

    # Phase I: loop over each model in the group and run an analysis function
    df_all = []
    ds_com = {}
    ds_ref = None
    for _, grp in comparison_data.groupby(ilamb3.conf["comparison_groupby"]):
        # Define what we will call the output artifacts
        source_name = "-".join(
            [grp.iloc[0][f] for f in ilamb3.conf["model_name_facets"]]
        )
        csv_file = output_path / f"{source_name}.csv"
        ref_file = output_path / "Reference.nc"
        com_file = output_path / f"{source_name}.nc"
        log_file = output_path / f"{source_name}.log"
        log_id = logger.add(log_file, backtrace=True, diagnose=True)

        # Attempt to load local assets if preferred
        if ilamb3.conf["use_cached_results"]:
            try:
                dfs, ds_ref, ds_com[source_name] = _load_local_assets(
                    csv_file, ref_file, com_file
                )
                df_all.append(dfs)
                logger.info(f"Using cached information {com_file}")
                continue
            except Exception:
                pass

        try:
            # Load data and run comparison
            ref = _load_reference_data(
                reference_data,
                variable,
                setup["sources"],
                setup["relationships"] if "relationships" in setup else {},
                transforms=transforms,
            )
            com = _load_comparison_data(
                grp,
                variable,
                alternate_vars=setup.get("alternate_vars", []),
                transforms=transforms,
            )
            dfs, ds_ref, ds_com[source_name] = run_analyses(ref, com, analyses)
            dfs["source"] = dfs["source"].str.replace("Comparison", source_name)

            # Write out artifacts
            dfs.to_csv(csv_file, index=False)
            if not ref_file.is_file():  # pragma: no cover
                ds_ref.to_netcdf(ref_file)
            ds_com[source_name].to_netcdf(com_file)
            df_all.append(dfs)
        except Exception:  # pragma: no cover
            logger.exception(
                f"ILAMB analysis '{block_name}' failed for '{source_name}'."
            )
            continue

        # Pop log and remove zero size files
        logger.remove(log_id)
        if log_file.stat().st_size == 0:  # pragma: no cover
            log_file.unlink()

    # Check that the reference intermediate data really was generated.
    if ds_ref is None:
        raise ValueError(
            "Reference intermediate data was not generated."
        )  # pragma: no cover

    # Phase 2: get plots and combine scalars and save
    plt.rcParams.update({"figure.max_open_warning": 0})
    df = pd.concat(df_all).drop_duplicates(
        subset=["source", "region", "analysis", "name"]
    )
    df = add_overall_score(df)
    df_plots = plot_analyses(df, ds_ref, ds_com, analyses, output_path)

    # Generate an output page
    if ilamb3.conf["debug_mode"] and (output_path / "index.html").is_file():
        return
    ds_ref.attrs["header"] = block_name
    html = generate_html_page(df, ds_ref, ds_com, df_plots)
    with open(output_path / "index.html", mode="w") as out:
        out.write(html)


def run_analyses(
    ref: xr.Dataset, com: xr.Dataset, analyses: dict[str, ILAMBAnalysis]
) -> tuple[pd.DataFrame, xr.Dataset, xr.Dataset]:
    """
    Run the input analyses on the given reference and comparison datasets.

    Parameters
    ----------
    ref : xr.Dataset
        The dataset which will be considered as the reference.
    com : xr.Dataset
        The dataset which will be considered as the comparison.
    analyses: dict[str, ILAMBAnalysis]
        A dictionary of analyses to

    Returns
    -------
    pd.DataFrame, xr.Dataset, xr.Dataset
        Analysis output, dataframe with scalar information and datasets with
        reference and comparison information for plotting.
    """
    dfs = []
    ds_refs = []
    ds_coms = []
    for aname, a in analyses.items():
        try:
            df, ds_ref, ds_com = a(ref, com)
        except AnalysisNotAppropriate:
            continue
        dfs.append(df)
        ds_refs.append(ds_ref)
        ds_coms.append(ds_com)
    dfs = pd.concat(dfs, ignore_index=True)
    dfs["name"] = dfs["name"] + " [" + dfs["units"] + "]"
    ds_ref = xr.merge(ds_refs)
    ds_com = xr.merge(ds_coms)
    return dfs, ds_ref, ds_com


def regenerate_figs(path: Path) -> bool:
    """
    Do we need to regenerate the figures?
    """
    path.mkdir(exist_ok=True, parents=True)
    png_files = list(path.glob("*.png"))
    if not png_files:
        return True
    first_png_time = min([p.stat().st_mtime for p in png_files])
    nc_files = list(path.glob("*.nc"))
    if not nc_files:
        return True
    last_nc_time = max([p.stat().st_mtime for p in nc_files])
    if last_nc_time > first_png_time:
        return True
    return False


def plot_analyses(
    df: pd.DataFrame,
    ref: xr.Dataset,
    com: dict[str, xr.Dataset],
    analyses: dict[str, ILAMBAnalysis],
    plot_path: Path,
) -> pd.DataFrame:
    """
    Plot analysis output encoded in each analysis.

    Parameters
    ----------
    df : pd.DataFrame
        A dataframe of all scalars from the analyses.
    ref : xr.Dataset
        A dataset containing reference data for plotting.
    com : dict[str,xr.Dataset]
        A dictionary of the comparison datasets whose keys are the model names.
    analyses : dict[str, ILAMBAnalysis]
        A dictionary of analyses to
    plot_path : Path
        A path to prepend all filenames.

    Returns
    -------
    pd.DataFrame
        A dataframe containing plot information and matplotlib axes.
    """
    if ilamb3.conf["debug_mode"] and not regenerate_figs(plot_path):
        return pd.DataFrame([])
    plot_path.mkdir(exist_ok=True, parents=True)
    df_plots = []
    for name, a in analyses.items():
        dfp = a.plots(
            df,
            ref,
            com,
        )
        for _, row in dfp.iterrows():
            if not row["axis"]:
                continue
            row["axis"].get_figure().savefig(
                plot_path / f"{row['source']}_{row['region']}_{row['name']}.png"
            )
        plt.close("all")
        if "analysis" not in dfp.columns:
            dfp["analysis"] = name
        df_plots.append(dfp)
    df_plots = pd.concat(df_plots)
    return df_plots


def generate_html_page(
    df: pd.DataFrame,
    ref: xr.Dataset,
    com: dict[str, xr.Dataset],
    df_plots: pd.DataFrame,
) -> str:
    """
    Generate an html page encoding all analysis data.

    Parameters
    ----------
    df : pd.DataFrame
        A dataframe of all scalars from the analyses.
    ref : xr.Dataset
        A dataset containing reference data for plotting.
    com : dict[str,xr.Dataset]
        A dictionary of the comparison datasets whose keys are the model names.
    df_plots : pd.DataFrame
        A dataframe containing plot information and matplotlib axes.

    Returns
    -------
    str
        The html page.
    """
    ilamb_regions = ilr.Regions()
    # Setup template analyses and plots
    analyses = {analysis: {} for analysis in df["analysis"].dropna().unique()}
    for (aname, pname), df_grp in df_plots.groupby(["analysis", "name"], sort=False):
        analyses[aname][pname] = []
        if len(df_grp["source"].unique()) == 1 and None in df_grp["source"].unique():
            analyses[aname][pname] += [{"None": f"None_RNAME_{pname}.png"}]
            continue
        if "Reference" in df_grp["source"].unique():
            analyses[aname][pname] += [{"Reference": f"Reference_RNAME_{pname}.png"}]
        analyses[aname][pname] += [{"Model": f"MNAME_RNAME_{pname}.png"}]
    ref_plots = list(df_plots[df_plots["source"] == "Reference"]["name"].unique())
    mod_plots = list(
        df_plots[~df_plots["source"].isin(["Reference", None])]["name"].unique()
    )
    all_plots = sorted(list(set(ref_plots) | set(mod_plots)))
    if not all_plots:
        all_plots = [""]

    # Setup template dictionary
    df = df.reset_index(drop=True)
    df["id"] = df.index
    data = {
        "page_header": ref.attrs["header"] if "header" in ref.attrs else "",
        "analysis_list": list(analyses.keys()),
        "model_names": [m for m in df["source"].unique() if m != "Reference"],
        "ref_plots": ref_plots,
        "mod_plots": mod_plots,
        "all_plots": all_plots,
        "regions": {
            (None if key == "None" else key): (
                "All Data" if key == "None" else ilamb_regions.get_name(key)
            )
            for key in df["region"].unique()
        },
        "analyses": analyses,
        "data_information": {
            key.capitalize(): ref.attrs[key]
            for key in ["title", "institution", "version", "doi"]
            if key in ref.attrs
        },
        "table_data": str(
            [row.to_dict() for _, row in df.drop(columns="units").iterrows()]
        ).replace("nan", "NaN"),
    }

    # Generate the html from the template
    template = importlib.resources.open_text(
        "ilamb3.templates", "dataset_page.html"
    ).read()
    html = Template(template).render(data)
    return html


def _flatten_dict(d: dict, parent_key: str = "", sep: str = "/") -> dict:
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict) and "sources" not in v:
            items.extend(_flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def _clean_pathname(filename: str) -> str:
    """Removes characters we do not want in our paths."""
    invalid_chars = r'[\\/:*?"<>|\s]'
    cleaned_filename = re.sub(invalid_chars, "", filename)
    return cleaned_filename


def _is_leaf(current: dict) -> bool:
    """Is the current item in the nested dictionary a leaf?"""
    if not isinstance(current, dict):
        return False
    if "sources" in current:
        return True
    return False


def _add_path(current: dict, path: Path | None = None) -> dict:
    """Recursively add the nested dictionary headings as a `path` in the leaves."""
    path = Path() if path is None else path
    for key, val in current.items():
        if not isinstance(val, dict):
            continue
        key_path = path / Path(_clean_pathname(key))
        if _is_leaf(val):
            val["path"] = str(key_path)
        else:
            current[key] = _add_path(val, key_path)
    return current


def _to_leaf_list(current: dict, leaf_list: list | None = None) -> list:
    """Recursively flatten the nested dictionary only returning the leaves."""
    leaf_list = [] if leaf_list is None else leaf_list
    for _, val in current.items():
        if not isinstance(val, dict):
            continue
        if _is_leaf(val):
            leaf_list.append(val)
        else:
            _to_leaf_list(val, leaf_list)
    return leaf_list


def _create_paths(current: dict, root: Path):
    """Recursively ensure paths in the leaves are created."""
    for _, val in current.items():
        if not isinstance(val, dict):
            continue
        if _is_leaf(val):
            if "path" in val:
                (root / Path(val["path"])).mkdir(parents=True, exist_ok=True)
        else:
            _create_paths(val, root)


def parse_benchmark_setup(yaml_file: str | Path) -> dict:
    """Parse the file which is analagous to the old configure file."""
    yaml_file = Path(yaml_file)
    with open(yaml_file) as fin:
        analyses = yaml.safe_load(fin)
    assert isinstance(analyses, dict)
    return analyses


def run_study(
    study_setup: str,
    df_datasets: pd.DataFrame,
    ref_datasets: pd.DataFrame | None = None,
    output_path: str | Path = "_build",
):
    ilamb3.conf["run_mode"] = "batch"
    output_path = Path(output_path)
    # Some yaml text that would get parsed like a dictionary.
    analyses = parse_benchmark_setup(study_setup)
    registry = analyses.pop("registry") if "registry" in analyses else "ilamb.txt"
    if registry == "ilamb.txt":
        reg = ilamb3.ilamb_catalog()
    elif registry == "iomb.txt":
        reg = ilamb3.iomb_catalog()
    else:
        raise ValueError("Unsupported registry.")

    # The yaml analysis setup can be as structured as the user needs. We are no longer
    # limited to the `h1` and `h2` headers from ILAMB 2.x. We will detect leaf nodes by
    # the presence of a `sources` dictionary.
    analyses = _add_path(analyses)

    # Various traversal actions
    _create_paths(analyses, output_path)

    # Create a list of just the leaves to use in creation all work combinations
    analyses_list = _to_leaf_list(analyses)

    # Run the confrontations
    for analysis in analyses_list:
        path = analysis.pop("path")
        try:
            run_single_block(
                path.split("/")[-1],
                (
                    ref_datasets
                    if ref_datasets is not None
                    else registry_to_dataframe(reg)
                ),
                df_datasets,
                output_path / path,
                **analysis,
            )
        except Exception:
            continue



"""
The ILAMB bias methodology.

See Also
--------
ILAMBAnalysis : The abstract base class from which this derives.
"""

from typing import Any

import numpy as np
import pandas as pd
import xarray as xr

import ilamb3.dataset as dset
import ilamb3.plot as plt
from ilamb3 import compare as cmp
from ilamb3.analysis.base import ILAMBAnalysis, integrate_or_mean, scalarify
from ilamb3.exceptions import AnalysisNotAppropriate


class cycle_analysis(ILAMBAnalysis):
    """
    The ILAMB annual cycle methodology.

    Parameters
    ----------
    required_variable : str
        The name of the variable to be used in this analysis.
    regions : list
        A list of region labels over which to apply the analysis.

    Methods
    -------
    required_variables
        What variables are required.
    __call__
        The method
    """

    def __init__(
        self,
        required_variable: str,
        regions: list[str | None] = [None],
        **kwargs: Any,  # this is so we can pass extra arguments without failure
    ):
        self.req_variable = required_variable
        self.regions = regions
        self.kwargs = kwargs

    def required_variables(self) -> list[str]:
        """
        Return the list of variables required for this analysis.

        Returns
        -------
        list
            The variable names used in this analysis.
        """
        return [self.req_variable]

    def __call__(
        self,
        ref: xr.Dataset,
        com: xr.Dataset,
    ) -> tuple[pd.DataFrame, xr.Dataset, xr.Dataset]:
        """
        Apply the ILAMB bias methodology on the given datasets.

        Parameters
        ----------
        ref : xr.Dataset
            The reference dataset.
        com : xr.Dataset
            The comparison dataset.

        Returns
        -------
        pd.DataFrame
            A dataframe with scalar and score information from the comparison.
        xr.Dataset
            A dataset containing reference grided information from the comparison.
        xr.Dataset
            A dataset containing comparison grided information from the comparison.
        """
        # Initialize
        analysis_name = "Annual Cycle"
        varname = self.req_variable

        # Make the variables comparable and force loading into memory
        ref, com = cmp.rename_dims(*cmp.make_comparable(ref, com, varname))

        # Is the time series long enough for this to be meaningful?
        if len(ref[dset.get_dim_name(ref, "time")]) < 12:
            raise AnalysisNotAppropriate()
        if len(com[dset.get_dim_name(com, "time")]) < 12:
            raise AnalysisNotAppropriate()

        # Compute the mean annual cycles
        ref = ref[varname].groupby("time.month").mean()
        com = com[varname].groupby("time.month").mean()

        # Get the timing of the maximum
        ref_tmax = xr.where(
            ref.notnull().any("month"), ref.fillna(0).argmax("month"), np.nan
        )
        com_tmax = xr.where(
            com.notnull().any("month"), com.fillna(0).argmax("month"), np.nan
        )

        # Compute the phase shift (difference in max month)
        ref_tmax_, com_tmax_ = cmp.nest_spatial_grids(ref_tmax, com_tmax)
        shift = com_tmax_ - ref_tmax_
        shift = xr.where(shift > 6, shift - 12, shift)
        shift = xr.where(shift < -6, shift + 12, shift)
        shift_score = 1 - np.abs(shift) / 6
        shift.attrs["units"] = "month"
        shift_score.attrs["units"] = "1"

        # Build output datasets
        ref_out = ref_tmax.to_dataset(name="tmax")
        com_out = shift.to_dataset(name="shift")
        com_out["cyclescore"] = shift_score
        try:
            lat_name = dset.get_dim_name(com_tmax, "lat")
            lon_name = dset.get_dim_name(com_tmax, "lon")
            com_tmax = com_tmax.rename({lat_name: "lat_", lon_name: "lon_"})
        except KeyError:
            pass
        com_out["tmax"] = com_tmax

        # Compute scalars over all regions
        dfs = []
        for region in self.regions:
            # Regional annual cycles
            ref_out[f"cycle_{region}"] = integrate_or_mean(
                ref, varname, region, mean=True
            )
            com_out[f"cycle_{region}"] = integrate_or_mean(
                com, varname, region, mean=True
            )

            # Regional mean phase shift
            val, unit = scalarify(com_out, "shift", region, mean=True)
            dfs.append(
                {
                    "source": "Comparison",
                    "region": str(region),
                    "analysis": analysis_name,
                    "name": "Phase Shift",
                    "type": "scalar",
                    "units": unit,
                    "value": val,
                }
            )

            # Regional cycle scores
            val, unit = scalarify(com_out, "cyclescore", region, mean=True)
            dfs.append(
                {
                    "source": "Comparison",
                    "region": str(region),
                    "analysis": analysis_name,
                    "name": "Seasonal Cycle Score",
                    "type": "score",
                    "units": unit,
                    "value": val,
                },
            )

        dfs = pd.DataFrame(dfs)
        return dfs, ref_out, com_out

    def plots(
        self,
        df: pd.DataFrame,
        ref: xr.Dataset,
        com: dict[str, xr.Dataset],
    ) -> pd.DataFrame:
        if "Annual Cycle" not in df["analysis"].unique():
            return pd.DataFrame()

        # Some initialization
        regions = [None if r == "None" else r for r in df["region"].unique()]
        com["Reference"] = ref

        # Setup plot data
        dfp = plt.determine_plot_limits(com).set_index("name")
        dfp.loc["tmax", ["cmap", "title", "low", "high"]] = [
            "rainbow",
            "Month of Maximum",
            -0.5,
            11.5,
        ]
        dfp.loc["shift", ["cmap", "title", "low", "high"]] = [
            "PRGn",
            "Phase Shift",
            -6,
            6,
        ]
        dfp.loc["cyclescore", ["cmap", "title"]] = ["plasma", "Cycle Score"]

        # Build up a dataframe of matplotlib axes
        plot_cbar_kwargs = {
            "tmax": {"ticks": range(12), "label": ""},
            "shift": {},
            "cyclescore": {},
        }
        axs = [
            {
                "name": plot,
                "title": dfp.loc[plot, "title"],
                "region": region,
                "source": source,
                "axis": (
                    plt.plot_map(
                        ds[plot],
                        region=region,
                        vmin=dfp.loc[plot, "low"],
                        vmax=dfp.loc[plot, "high"],
                        cmap=dfp.loc[plot, "cmap"],
                        title=source + " " + dfp.loc[plot, "title"],
                        ncolors=12 if plot == "tmax" else 9,
                        ticklabels=(
                            [
                                "Jan",
                                "Feb",
                                "Mar",
                                "Apr",
                                "May",
                                "Jun",
                                "Jul",
                                "Aug",
                                "Sep",
                                "Oct",
                                "Nov",
                                "Dec",
                            ]
                            if plot == "tmax"
                            else None
                        ),
                        **{"cbar_kwargs": plot_cbar_kwargs[plot]},
                    )
                    if plot in ds
                    else pd.NA
                ),
            }
            for plot in ["tmax", "shift", "cyclescore"]
            for source, ds in com.items()
            for region in regions
        ]

        axs += [
            {
                "name": "cycle",
                "title": "Annual Cycle",
                "region": plot.split("_")[-1],
                "source": source,
                "axis": (
                    plt.plot_curve(
                        {source: ds} | {"Reference": ref},
                        plot,
                        vmin=dfp.loc[plot, "low"]
                        - 0.05 * (dfp.loc[plot, "high"] - dfp.loc[plot, "low"]),
                        vmax=dfp.loc[plot, "high"]
                        + 0.05 * (dfp.loc[plot, "high"] - dfp.loc[plot, "low"]),
                        title=f"{source} Annual Cycle",
                        xticks=range(1, 13),
                        xticklabels=[
                            "Jan",
                            "Feb",
                            "Mar",
                            "Apr",
                            "May",
                            "Jun",
                            "Jul",
                            "Aug",
                            "Sep",
                            "Oct",
                            "Nov",
                            "Dec",
                        ],
                    )
                    if plot in ds
                    else pd.NA
                ),
            }
            for plot in [f"cycle_{region}" for region in regions]
            for source, ds in com.items()
            if source != "Reference"
        ]

        axs = pd.DataFrame(axs).dropna(subset=["axis"])
        return axs



"""
The ILAMB bias methodology.

See Also
--------
ILAMBAnalysis : The abstract base class from which this derives.
"""

from typing import Any

import numpy as np
import pandas as pd
import xarray as xr

import ilamb3.plot as ilplt
import ilamb3.regions as ilr
from ilamb3 import compare as cmp
from ilamb3 import dataset as dset
from ilamb3.analysis.base import ILAMBAnalysis


class spatial_distribution_analysis(ILAMBAnalysis):
    """
    The ILAMB bias methodology.

    Parameters
    ----------
    required_variable : str
        The name of the variable to be used in this analysis.
    regions : list
        A list of region labels over which to apply the analysis.

    Methods
    -------
    required_variables
        What variables are required.
    __call__
        The method
    """

    def __init__(
        self,
        required_variable: str,
        regions: list[str | None] = [None],
        **kwargs: Any,  # this is so we can pass extra arguments without failure
    ):
        self.req_variable = required_variable
        self.regions = regions
        self.kwargs = kwargs

    def required_variables(self) -> list[str]:
        """
        Return the list of variables required for this analysis.

        Returns
        -------
        list
            The variable names used in this analysis.
        """
        return [self.req_variable]

    def __call__(
        self,
        ref: xr.Dataset,
        com: xr.Dataset,
    ) -> tuple[pd.DataFrame, xr.Dataset, xr.Dataset]:
        """
        Apply the ILAMB bias methodology on the given datasets.

        Parameters
        ----------
        ref : xr.Dataset
            The reference dataset.
        com : xr.Dataset
            The comparison dataset.

        Returns
        -------
        pd.DataFrame
            A dataframe with scalar and score information from the comparison.
        xr.Dataset
            A dataset containing reference grided information from the comparison.
        xr.Dataset
            A dataset containing comparison grided information from the comparison.
        """
        # Initialize
        analysis_name = "Spatial Distribution"
        varname = self.req_variable

        # Make the variables comparable and force loading into memory
        ref, com = cmp.make_comparable(ref, com, varname)

        # Temporal means across the time period...
        ref = (
            dset.integrate_time(ref, varname, mean=True)
            if dset.is_temporal(ref[varname])
            else ref[varname]
        )
        com = (
            dset.integrate_time(com, varname, mean=True)
            if dset.is_temporal(com[varname])
            else com[varname]
        )

        # ... on the same grid
        ref, com = cmp.rename_dims(*cmp.nest_spatial_grids(ref, com))

        # Compute scalars over all regions
        dfs = []
        ilamb_regions = ilr.Regions()
        for region in self.regions:
            # Get regional versions
            rref = ilamb_regions.restrict_to_region(ref, region)
            rcom = ilamb_regions.restrict_to_region(com, region)

            # Spatial standard deviation
            ref_std = float(rref.std())
            com_std = float(rcom.std())
            norm_std = com_std / ref_std

            # Correlation
            isnan = rref.isnull() | rcom.isnull()
            rref = np.ma.masked_invalid(
                xr.where(isnan, np.nan, rref).values
            ).compressed()
            rcom = np.ma.masked_invalid(
                xr.where(isnan, np.nan, rcom).values
            ).compressed()
            corr = float(np.corrcoef(rref, rcom)[0, 1])
            taylor_score = 4 * (1 + corr) / ((norm_std + 1 / norm_std) ** 2 * 2)

            dfs += [
                {
                    "source": "Comparison",
                    "region": str(region),
                    "analysis": analysis_name,
                    "name": "Normalized Standard Deviation",
                    "type": "scalar",
                    "units": "1",
                    "value": norm_std,
                },
                {
                    "source": "Comparison",
                    "region": str(region),
                    "analysis": analysis_name,
                    "name": "Correlation",
                    "type": "scalar",
                    "units": "1",
                    "value": corr,
                },
                {
                    "source": "Comparison",
                    "region": str(region),
                    "analysis": analysis_name,
                    "name": "Spatial Distribution Score",
                    "type": "score",
                    "units": "1",
                    "value": taylor_score,
                },
            ]

        dfs = pd.DataFrame(dfs)
        return dfs, xr.Dataset(), xr.Dataset()

    def plots(
        self,
        df: pd.DataFrame,
        ref: xr.Dataset,
        com: dict[str, xr.Dataset],
    ) -> pd.DataFrame:
        # Add the Taylor diagram
        axs = []
        for region in df["region"].unique():
            axs += [
                {
                    "name": "taylor",
                    "title": "Spatial Distribution",
                    "region": str(region),
                    "source": None,
                    "axis": ilplt.plot_taylor_diagram(df[df["region"] == region]),
                }
            ]
        axs = pd.DataFrame(axs).dropna(subset=["axis"])
        return axs
